(function () {

/* Imports */
var Meteor = Package.meteor.Meteor;
var MongoInternals = Package.mongo.MongoInternals;
var Mongo = Package.mongo.Mongo;
var Random = Package.random.Random;
var check = Package.check.check;
var Match = Package.check.Match;
var _ = Package.underscore._;
var Accounts = Package['accounts-base'].Accounts;

/* Package-scope variables */
var SandstormDb, Packages, DevApps, UserActions, Grains, RoleAssignments, Contacts, Sessions, SignupKeys, ActivityStats, DeleteStats, FileTokens, ApiTokens, Notifications, StatsTokens, Misc, Settings, Migrations, isDemoUser, isSignedUp, isSignedUpOrDemo, isUserOverQuota, isUserExcessivelyOverQuota, isAdmin, isAdminById, findAdminUserForToken, matchWildcardHost, makeWildcardHost, getWildcardOrigin, allowDevAccounts, roleAssignmentPattern;

(function () {

////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                    //
// packages/sandstorm-db/db.js                                                                        //
//                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                      //
// Sandstorm - Personal Cloud Sandbox                                                                 // 1
// Copyright (c) 2014 Sandstorm Development Group, Inc. and contributors                              // 2
// All rights reserved.                                                                               // 3
//                                                                                                    // 4
// Licensed under the Apache License, Version 2.0 (the "License");                                    // 5
// you may not use this file except in compliance with the License.                                   // 6
// You may obtain a copy of the License at                                                            // 7
//                                                                                                    // 8
//   http://www.apache.org/licenses/LICENSE-2.0                                                       // 9
//                                                                                                    // 10
// Unless required by applicable law or agreed to in writing, software                                // 11
// distributed under the License is distributed on an "AS IS" BASIS,                                  // 12
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.                           // 13
// See the License for the specific language governing permissions and                                // 14
// limitations under the License.                                                                     // 15
                                                                                                      // 16
// This file defines the database schema.                                                             // 17
                                                                                                      // 18
Packages = new Mongo.Collection("packages");                                                          // 19
// Packages which are installed or downloading.                                                       // 20
//                                                                                                    // 21
// Each contains:                                                                                     // 22
//   _id:  128-bit prefix of SHA-256 hash of spk file, hex-encoded.                                   // 23
//   status:  String.  One of "download", "verify", "unpack", "analyze", "ready", "failed", "delete"  // 24
//   progress:  Float.  -1 = N/A, 0-1 = fractional progress (e.g. download percentage),               // 25
//       >1 = download byte count.                                                                    // 26
//   error:  If status is "failed", error message string.                                             // 27
//   manifest:  If status is "ready", the package manifest.  See "Manifest" in grain.capnp.           // 28
//   appId:  If status is "ready", the application ID string.  Packages representing different        // 29
//       versions of the same app have the same appId.  The spk tool defines the app ID format        // 30
//       and can cryptographically verify that a package belongs to a particular app ID.              // 31
                                                                                                      // 32
DevApps = new Mongo.Collection("devapps");                                                            // 33
// List of applications currently made available via the dev tools running on the local machine.      // 34
// This is normally empty; the only time it is non-empty is when a developer is using the spk tool    // 35
// on the local machine to publish an under-development app to this server. That should only ever     // 36
// happen on developers' desktop machines.                                                            // 37
//                                                                                                    // 38
// While a dev app is published, it automatically appears as installed by every user of the server,   // 39
// and it overrides all packages with the same application ID. If any instances of those packages     // 40
// are currently open, they are killed and reset on publish.                                          // 41
//                                                                                                    // 42
// When the dev tool disconnects, the app is automatically unpublished, and any open instances        // 43
// are again killed and refreshed.                                                                    // 44
//                                                                                                    // 45
// Each contains:                                                                                     // 46
//   _id:  The application ID string (as with Packages.appId).                                        // 47
//   packageId:  The directory name where the dev package is mounted.                                 // 48
//   timestamp:  Time when the package was last updated. If this changes while the package is         // 49
//     published, all running instances are reset. This is used e.g. to reset the app each time       // 50
//     changes are made to the source code.                                                           // 51
//   manifest:  The app's manifest, as with Packages.manifest.                                        // 52
                                                                                                      // 53
UserActions = new Mongo.Collection("userActions");                                                    // 54
// List of actions that each user has installed which create new grains.  Each app may install        // 55
// some number of actions (usually, one).                                                             // 56
//                                                                                                    // 57
// Each contains:                                                                                     // 58
//   _id:  random                                                                                     // 59
//   userId:  User who has installed this action.                                                     // 60
//   packageId:  Package used to run this action.                                                     // 61
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.               // 62
//   appTitle:  Same as Packages.findOne(packageId).manifest.appTitle.defaultText; denormalized so    // 63
//       that clients can access it without subscribing to the Packages collection.                   // 64
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for           // 65
//       searchability.                                                                               // 66
//   appMarketingVersion:  Human-readable presentation of the app version, e.g. "2.9.17"              // 67
//   title:  Human-readable title for this action, e.g. "New Spreadsheet".                            // 68
//   command:  Manifest.Command to run this action (see package.capnp).                               // 69
                                                                                                      // 70
Grains = new Mongo.Collection("grains");                                                              // 71
// Grains belonging to users.                                                                         // 72
//                                                                                                    // 73
// Each contains:                                                                                     // 74
//   _id:  random                                                                                     // 75
//   packageId:  _id of the package of which this grain is an instance.                               // 76
//   appId:  Same as Packages.findOne(packageId).appId; denormalized for searchability.               // 77
//   appVersion:  Same as Packages.findOne(packageId).manifest.appVersion; denormalized for           // 78
//       searchability.                                                                               // 79
//   userId:  User who owns this grain.                                                               // 80
//   title:  Human-readable string title, as chosen by the user.                                      // 81
//   lastUsed:  Date when the grain was last used by a user.                                          // 82
//   private: If true, then knowledge of `_id` does not suffice to open this grain.                   // 83
//   cachedViewInfo: The JSON-encoded result of `UiView.getViewInfo()`, cached from the most recent   // 84
//                   time a session to this grain was opened.                                         // 85
//                                                                                                    // 86
// The following fields *might* also exist. These are temporary hacks used to implement e-mail and    // 87
// web publishing functionality without powerbox support; they will be replaced once the powerbox     // 88
// is implemented.                                                                                    // 89
//   publicId:  An id used to publicly identify this grain. Used e.g. to route incoming e-mail and    // 90
//       web publishing. This field is initialized when first requested by the app.                   // 91
                                                                                                      // 92
RoleAssignments = new Mongo.Collection("roleAssignments");                                            // 93
// *OBSOLETE* Before `user` was a variant of ApiTokenOwner, this collection was used to store edges   // 94
// in the permissions sharing graph. This functionality has been subsumed by the ApiTokens            // 95
// collection.                                                                                        // 96
                                                                                                      // 97
Contacts = new Mongo.Collection("contacts");                                                          // 98
// Edges in the social graph.                                                                         // 99
//                                                                                                    // 100
// If Alice has Bob as a contact, then she is allowed to see Bob's profile information and Bob        // 101
// will show up in her user-picker UI for actions like share-by-identity.                             // 102
//                                                                                                    // 103
// Contacts are not symmetric. Bob might be one of Alice's contacts even if Alice is not one of       // 104
// Bob's.                                                                                             // 105
//                                                                                                    // 106
// Each contains:                                                                                     // 107
//   _id: random                                                                                      // 108
//   ownerId: The `_id` of the user who owns this contact.                                            // 109
//   userId:  The `_id` of the contacted user.                                                        // 110
//   petname: Human-readable label chosen by and only visible to the owner. Uniquely identifies       // 111
//            the contact to the owner.                                                               // 112
//   created: Date when this contact was created.                                                     // 113
                                                                                                      // 114
Sessions = new Mongo.Collection("sessions");                                                          // 115
// UI sessions open to particular grains.  A new session is created each time a user opens a grain.   // 116
//                                                                                                    // 117
// Each contains:                                                                                     // 118
//   _id:  random                                                                                     // 119
//   grainId:  _id of the grain to which this session is connected.                                   // 120
//   hostId: ID part of the hostname from which this grain is being served. I.e. this replaces the    // 121
//       '*' in WILDCARD_HOST.                                                                        // 122
//   timestamp:  Time of last keep-alive message to this session.  Sessions time out after some       // 123
//       period.                                                                                      // 124
//   userId:  User who owns this session.                                                             // 125
//   hashedToken: If the session is owned by an anonymous user, the _id of the entry in ApiTokens     // 126
//       that was used to open it. Note that for old-style sharing (i.e. when !grain.private),        // 127
//       anonymous users can get access without an API token and so neither userId nor hashedToken    // 128
//       are present.                                                                                 // 129
//   powerboxView: If present, this is a view that should be presented as part of a powerbox          // 130
//       interaction.                                                                                 // 131
//     offer: The webkey that corresponds to cap that was passed to the `offer` RPC.                  // 132
                                                                                                      // 133
SignupKeys = new Mongo.Collection("signupKeys");                                                      // 134
// Invite keys which may be used by users to get access to Sandstorm.                                 // 135
//                                                                                                    // 136
// Each contains:                                                                                     // 137
//   _id:  random                                                                                     // 138
//   used:  Boolean indicating whether this key has already been consumed.                            // 139
//   note:  Text note assigned when creating key, to keep track of e.g. whom the key was for.         // 140
//   email: If this key was sent as an email invite, the email address to which it was sent.          // 141
//   quota: If present, the storage quota to assign to the user claiming this invite.                 // 142
                                                                                                      // 143
ActivityStats = new Mongo.Collection("activityStats");                                                // 144
// Contains usage statistics taken on a regular interval. Each entry is a data point.                 // 145
//                                                                                                    // 146
// Each contains:                                                                                     // 147
//   timestamp: Date when measurements were taken.                                                    // 148
//   daily: Contains stats counts pertaining to the last day before the sample time.                  // 149
//   weekly: Contains stats counts pertaining to the last seven days before the sample time.          // 150
//   monthly: Contains stats counts pertaining to the last thirty days before the timestamp.          // 151
//                                                                                                    // 152
// Each of daily, weekly, and monthly contains:                                                       // 153
//   activeUsers: The number of unique users who have used a grain on the server in the time          // 154
//       interval. Only counts logged-in users.                                                       // 155
//   activeGrains: The number of unique grains that have been used in the time interval.              // 156
                                                                                                      // 157
DeleteStats = new Mongo.Collection("deleteStats");                                                    // 158
// Contains records of objects that were deleted, for stat-keeping purposes.                          // 159
//                                                                                                    // 160
// Each contains:                                                                                     // 161
//   type: "grain" or "user" or "appDemoUser"                                                         // 162
//   lastActive: Date of the user's or grain's last activity.                                         // 163
                                                                                                      // 164
FileTokens = new Mongo.Collection("fileTokens");                                                      // 165
// Tokens corresponding to backup files that are currently stored on the server. A user receives      // 166
// a token when they create a backup file (either by uploading it, or by backing up one of their      // 167
// grains) and may use the token to read the file (either to download it, or to restore a new         // 168
// grain from it).                                                                                    // 169
//                                                                                                    // 170
// Each contains:                                                                                     // 171
//   _id:       The unguessable token string.                                                         // 172
//   name:      Suggested filename.                                                                   // 173
//   timestamp: File creation time. Used to figure out when the token and file should be wiped.       // 174
                                                                                                      // 175
ApiTokens = new Mongo.Collection("apiTokens");                                                        // 176
// Access tokens for APIs exported by apps.                                                           // 177
//                                                                                                    // 178
// Originally API tokens were only used by external users through the HTTP API endpoint. However,     // 179
// now they are also used to implement SturdyRefs, not just held by external users, but also when     // 180
// an app holds a SturdyRef to another app within the same server. See the various `save()`,          // 181
// `restore()`, and `drop()` methods in `grain.capnp` (on `SansdtormApi`, `AppPersistent`, and        // 182
// `MainView`) -- the fields of type `Data` are API tokens.                                           // 183
//                                                                                                    // 184
// Each contains:                                                                                     // 185
//   _id:       A SHA-256 hash of the token.                                                          // 186
//   grainId:   The grain servicing this API. (Not present if the API isn't serviced by a grain.)     // 187
//   userId:    For UiView capabilities, this is the user for whom the view is attenuated. That       // 188
//              is, the UiView's newSession() method will intersect the requested permissions with    // 189
//              this user's permissions before forwarding on to the underlying app. If `userId` is    // 190
//              not present, then no user attenuation is applied, i.e. this is a raw UiView as        // 191
//              implemented by the app. (The `roleAssignment` field, below, may still apply.)         // 192
//              For non-UiView capabilities, `userId` is never present.                               // 193
//              Note that this is NOT the user against whom the `requiredPermissions` parameter of    // 194
//              `SandstormApi.restore()` is checked; that would be `owner.grain.introducerUser`.      // 195
//   userInfo:  *DEPRECATED* For API tokens created by the app through HackSessionContext, the        // 196
//              UserInfo struct that should be passed to `newSession()` when exercising this token,   // 197
//              in decoded (JS object) format. This is a temporary hack. `userId` is never present    // 198
//              when `userInfo` is present.                                                           // 199
//   roleAssignment: If this API token represents a UiView, this field contains a JSON-encoded        // 200
//              Grain.ViewSharingLink.RoleAssignment representing the permissions it carries. These   // 201
//              permissions will be intersected with those held by `userId` when the view is opened.  // 202
//   forSharing: If true, requests sent to the HTTP API endpoint with this token will be treated as   // 203
//              anonymous rather than as directly associated with `userId`. This has no effect on     // 204
//              the permissions granted.                                                              // 205
//   objectId:  If present, this token represents an arbitrary Cap'n Proto capability exported by     // 206
//              the app or its supervisor (whereas without this it strictly represents UiView).       // 207
//              sturdyRef is the JSON-encoded SupervisorObjectId (defined in `supervisor.capnp`).     // 208
//              Note that if the SupervisorObjectId contains an AppObjectId, that field is            // 209
//              treated as type AnyPointer, and so encoded as a raw Cap'n Proto message.              // 210
//   frontendRef: If present, this token actually refers to an object implemented by the front-end,   // 211
//              not a particular grain. (`grainId` and `userId` are not set.) This is an object       // 212
//              containing exactly one of the following fields:                                       // 213
//       notificationHandle: A `Handle` for an ongoing notification, as returned by                   // 214
//                           `NotificationTarget.addOngoing`. The value is an `_id` from the          // 215
//                           `Notifications` collection.                                              // 216
//       ipNetwork: An IpNetwork capability that is implemented by the frontend. Eventually, this     // 217
//                  will be moved out of the frontend and into the backend, but we'll migrate the     // 218
//                  database when that happens. This field contains the boolean true to signify that  // 219
//                  it has been set.                                                                  // 220
//       ipInterface: Ditto IpNetwork, except it's an IpInterface.                                    // 221
//   parentToken: If present, then this token represents exactly the capability represented by        // 222
//              the ApiToken with _id = parentToken, except possibly (if it is a UiView) attenuated   // 223
//              by `roleAssignment` (if present). To facilitate permissions computations, if the      // 224
//              capability is a UiView, then `grainId` is set to the backing grain and `userId` is    // 225
//              set to the user who shared the view. None of `userInfo`, `objectId`, or               // 226
//              `frontendRef` are present when `parentToken` is present.                              // 227
//   petname:   Human-readable label for this access token, useful for identifying tokens for         // 228
//              revocation. This should be displayed when visualizing incoming capabilities to        // 229
//              the grain identified by `grainId`.                                                    // 230
//   created:   Date when this token was created.                                                     // 231
//   revoked:   If true, then this sturdyref has been revoked and can no longer be restored. It may   // 232
//              become un-revoked in the future.                                                      // 233
//   expires:   Optional expiration Date. If undefined, the token does not expire.                    // 234
//   owner:     A `ApiTokenRefOwner` (defined in `supervisor.capnp`, stored as a JSON object)         // 235
//              as passed to the `save()` call that created this token. If not present, treat         // 236
//              as `webkey` (the default for `ApiTokenOwner`).                                        // 237
//   expiresIfUnused:                                                                                 // 238
//              Optional Date after which the token, if it has not been used yet, expires.            // 239
//              This field should be cleared on a token's first use.                                  // 240
//   requirements: List of conditions which must hold for this token to be considered valid.          // 241
//              Semantically, this list specifies the powers which were *used* to originally          // 242
//              create the token. If any condition in the list becomes untrue, then the token must    // 243
//              be considered revoked, and all live refs and sturdy refs obtained transitively        // 244
//              through it must also become revoked. Each item is the JSON serialization of the       // 245
//              `MembraneRequirement` structure defined in `supervisor.capnp`.                        // 246
//                                                                                                    // 247
// It is important to note that a token's owner and provider are independent from each other. To      // 248
// illustrate, here is an approximate definition of ApiToken in pseudo Cap'n Proto schema language:   // 249
//                                                                                                    // 250
// struct ApiToken {                                                                                  // 251
//   owner :ApiTokenOwner;                                                                            // 252
//   provider :union {                                                                                // 253
//     grain :group {                                                                                 // 254
//       grainId :Text;                                                                               // 255
//       union {                                                                                      // 256
//         uiView :group {                                                                            // 257
//           userId :Text;                                                                            // 258
//           roleAssignment :RoleAssignment;                                                          // 259
//           forSharing :Bool;                                                                        // 260
//         }                                                                                          // 261
//         objectId :SupervisorObjectId;                                                              // 262
//       }                                                                                            // 263
//     }                                                                                              // 264
//     frontendRef :union {                                                                           // 265
//        notificationHandle :Text;                                                                   // 266
//        ipNetwork :Bool;                                                                            // 267
//        ipInterface :Bool;                                                                          // 268
//     }                                                                                              // 269
//     child :group {                                                                                 // 270
//       parentToken :Text;                                                                           // 271
//       union {                                                                                      // 272
//         uiView :group {                                                                            // 273
//           grainId :Text;                                                                           // 274
//           userId :Text;                                                                            // 275
//           roleAssignment :RoleAssignment = (allAccess = ());                                       // 276
//         }                                                                                          // 277
//         other :Void;                                                                               // 278
//       }                                                                                            // 279
//     }                                                                                              // 280
//   }                                                                                                // 281
//   requirements: List(Supervisor.MembraneRequirement);                                              // 282
//   ...                                                                                              // 283
// }                                                                                                  // 284
                                                                                                      // 285
Notifications = new Mongo.Collection("notifications");                                                // 286
// Notifications for a user.                                                                          // 287
//                                                                                                    // 288
// Each contains:                                                                                     // 289
//   _id:          random                                                                             // 290
//   ongoing:      If present, this is an ongoing notification, and this field contains an            // 291
//                 ApiToken referencing the `OngoingNotification` capability.                         // 292
//   grainId:      The grain originating this notification, if any.                                   // 293
//   userId:       The user receiving the notification.                                               // 294
//   text:         The JSON-ified LocalizedText to display in the notification.                       // 295
//   isUnread:     Boolean indicating if this notification is unread.                                 // 296
//   timestamp:    Date when this notification was last updated                                       // 297
                                                                                                      // 298
StatsTokens = new Mongo.Collection("statsTokens");                                                    // 299
// Access tokens for the Stats collection                                                             // 300
//                                                                                                    // 301
// These tokens are used for accessing the ActivityStats collection remotely                          // 302
// (ie. from a dashboard webapp)                                                                      // 303
//                                                                                                    // 304
// Each contains:                                                                                     // 305
//   _id:       The token. At least 128 bits entropy (Random.id(22)).                                 // 306
                                                                                                      // 307
Misc = new Mongo.Collection("misc");                                                                  // 308
// Miscellaneous configuration and other settings                                                     // 309
//                                                                                                    // 310
// This table is currently only used for persisting BASE_URL from one session to the next,            // 311
// but in general any miscellaneous settings should go in here                                        // 312
//                                                                                                    // 313
// Each contains:                                                                                     // 314
//   _id:       The name of the setting. eg. "BASE_URL"                                               // 315
//   value:     The value of the setting.                                                             // 316
                                                                                                      // 317
Settings = new Mongo.Collection("settings");                                                          // 318
// Settings for this Sandstorm instance go here. They are configured through the adminSettings        // 319
// route. This collection differs from misc in that this collection is completely user controlled.    // 320
//                                                                                                    // 321
// Each contains:                                                                                     // 322
//   _id:       The name of the setting. eg. "MAIL_URL"                                               // 323
//   value:     The value of the setting.                                                             // 324
//   potentially other fields that are unique to the setting                                          // 325
                                                                                                      // 326
Migrations = new Mongo.Collection("migrations");                                                      // 327
// This table tracks which migrations we have applied to this instance.                               // 328
// It contains a single entry:                                                                        // 329
//   _id:       "migrations_applied"                                                                  // 330
//   value:     The number of migrations this instance has successfully completed.                    // 331
                                                                                                      // 332
if (Meteor.isServer) {                                                                                // 333
  Meteor.publish("credentials", function () {                                                         // 334
    // Data needed for isSignedUp() and isAdmin() to work.                                            // 335
                                                                                                      // 336
    if (this.userId) {                                                                                // 337
      return Meteor.users.find({_id: this.userId},                                                    // 338
          {fields: {signupKey: 1, isAdmin: 1, expires: 1, quota: 1, storageUsage: 1}});               // 339
    } else {                                                                                          // 340
      return [];                                                                                      // 341
    }                                                                                                 // 342
  });                                                                                                 // 343
                                                                                                      // 344
  // The first user to sign in should be automatically upgraded to admin.                             // 345
  Accounts.onCreateUser(function (options, user) {                                                    // 346
    // Dev users are identified by having the devName field                                           // 347
    // Don't count them in our find and don't give them admin                                         // 348
    if (Meteor.users.find({devName: {$exists: 0}}).count() === 0 && !user.devName) {                  // 349
      user.isAdmin = true;                                                                            // 350
      user.signupKey = "admin";                                                                       // 351
    }                                                                                                 // 352
                                                                                                      // 353
    if (options.profile) {                                                                            // 354
      user.profile = options.profile;                                                                 // 355
    }                                                                                                 // 356
                                                                                                      // 357
    return user;                                                                                      // 358
  });                                                                                                 // 359
}                                                                                                     // 360
                                                                                                      // 361
isDemoUser = function() {                                                                             // 362
  // Returns true if this is a demo user.                                                             // 363
                                                                                                      // 364
  var user = Meteor.user();                                                                           // 365
  if (user && user.expires) {                                                                         // 366
    return true;                                                                                      // 367
  } else {                                                                                            // 368
    return false;                                                                                     // 369
  }                                                                                                   // 370
}                                                                                                     // 371
                                                                                                      // 372
isSignedUp = function() {                                                                             // 373
  // Returns true if the user has presented an invite key.                                            // 374
                                                                                                      // 375
  var user = Meteor.user();                                                                           // 376
  if (user && user.signupKey) {                                                                       // 377
    return true;                                                                                      // 378
  } else {                                                                                            // 379
    return false;                                                                                     // 380
  }                                                                                                   // 381
}                                                                                                     // 382
                                                                                                      // 383
isSignedUpOrDemo = function () {                                                                      // 384
  var user = Meteor.user();                                                                           // 385
  if (user && (user.signupKey || user.expires)) {                                                     // 386
    return true;                                                                                      // 387
  } else {                                                                                            // 388
    return false;                                                                                     // 389
  }                                                                                                   // 390
}                                                                                                     // 391
                                                                                                      // 392
isUserOverQuota = function (user) {                                                                   // 393
  return Meteor.settings.public.quotaEnabled &&                                                       // 394
         typeof user.quota === "number" &&                                                            // 395
         user.storageUsage && user.storageUsage >= user.quota;                                        // 396
}                                                                                                     // 397
                                                                                                      // 398
isUserExcessivelyOverQuota = function (user) {                                                        // 399
  return Meteor.settings.public.quotaEnabled &&                                                       // 400
         typeof user.quota === "number" &&                                                            // 401
         user.storageUsage && user.storageUsage >= user.quota * 1.2;                                  // 402
}                                                                                                     // 403
                                                                                                      // 404
isAdmin = function() {                                                                                // 405
  // Returns true if the user is the administrator.                                                   // 406
                                                                                                      // 407
  var user = Meteor.user();                                                                           // 408
  if (user && user.isAdmin) {                                                                         // 409
    return true;                                                                                      // 410
  } else {                                                                                            // 411
    return false;                                                                                     // 412
  }                                                                                                   // 413
}                                                                                                     // 414
                                                                                                      // 415
isAdminById = function(id) {                                                                          // 416
  // Returns true if the user's id is the administrator.                                              // 417
                                                                                                      // 418
  var user = Meteor.users.findOne({_id: id}, {fields: {isAdmin: 1}})                                  // 419
  if (user && user.isAdmin) {                                                                         // 420
    return true;                                                                                      // 421
  } else {                                                                                            // 422
    return false;                                                                                     // 423
  }                                                                                                   // 424
}                                                                                                     // 425
                                                                                                      // 426
findAdminUserForToken = function (token) {                                                            // 427
  if (!token.requirements) {                                                                          // 428
    return;                                                                                           // 429
  }                                                                                                   // 430
  var requirements = token.requirements.filter(function (requirement) {                               // 431
    return "userIsAdmin" in requirement;                                                              // 432
  });                                                                                                 // 433
                                                                                                      // 434
  if (requirements.length > 1) {                                                                      // 435
    return;                                                                                           // 436
  }                                                                                                   // 437
  if (requirements.length === 0) {                                                                    // 438
    return;                                                                                           // 439
  }                                                                                                   // 440
  return requirements[0].userIsAdmin;                                                                 // 441
};                                                                                                    // 442
                                                                                                      // 443
var wildcardHost = Meteor.settings.public.wildcardHost.toLowerCase().split("*");                      // 444
                                                                                                      // 445
if (wildcardHost.length != 2) {                                                                       // 446
  throw new Error("Wildcard host must contain exactly one asterisk.");                                // 447
}                                                                                                     // 448
                                                                                                      // 449
matchWildcardHost = function(host) {                                                                  // 450
  // See if the hostname is a member of our wildcard. If so, extract the ID.                          // 451
                                                                                                      // 452
  var prefix = wildcardHost[0];                                                                       // 453
  var suffix = wildcardHost[1];                                                                       // 454
  if (host.lastIndexOf(prefix, 0) >= 0 &&                                                             // 455
      host.indexOf(suffix, -suffix.length) >= 0 &&                                                    // 456
      host.length >= prefix.length + suffix.length) {                                                 // 457
    var id = host.slice(prefix.length, -suffix.length);                                               // 458
    if (id.match(/^[a-z0-9]*$/)) {                                                                    // 459
      return id;                                                                                      // 460
    }                                                                                                 // 461
  }                                                                                                   // 462
                                                                                                      // 463
  return null;                                                                                        // 464
}                                                                                                     // 465
                                                                                                      // 466
makeWildcardHost = function (id) {                                                                    // 467
  return wildcardHost[0] + id + wildcardHost[1];                                                      // 468
}                                                                                                     // 469
                                                                                                      // 470
if (Meteor.isServer) {                                                                                // 471
  var Url = Npm.require("url");                                                                       // 472
  getWildcardOrigin = function () {                                                                   // 473
    // The wildcard URL can be something like "foo-*-bar.example.com", but sometimes when we're       // 474
    // trying to specify a pattern matching hostnames (say, a Content-Security-Policy directive),     // 475
    // an astrisk is only allowed as the first character and must be followed by a period. So we need // 476
    // "*.example.com" instead -- which matches more than we actually want, but is the best we can    // 477
    // really do. We also add the protocol to the front (again, that's what CSP wants).               // 478
                                                                                                      // 479
    // TODO(cleanup): `protocol` is computed in other files, like proxy.js. Put it somewhere common.  // 480
    var protocol = Url.parse(process.env.ROOT_URL).protocol;                                          // 481
                                                                                                      // 482
    var dotPos = wildcardHost[1].indexOf(".");                                                        // 483
    if (dotPos < 0) {                                                                                 // 484
      return protocol + "//*";                                                                        // 485
    } else {                                                                                          // 486
      return protocol + "//*" + wildcardHost[1].slice(dotPos);                                        // 487
    }                                                                                                 // 488
  }                                                                                                   // 489
}                                                                                                     // 490
                                                                                                      // 491
allowDevAccounts = function () {                                                                      // 492
  var setting = Settings.findOne({_id: "devAccounts"});                                               // 493
  if (setting) {                                                                                      // 494
    return setting.value;                                                                             // 495
  } else {                                                                                            // 496
    return Meteor.settings && Meteor.settings.public &&                                               // 497
           Meteor.settings.public.allowDevAccounts;                                                   // 498
  }                                                                                                   // 499
};                                                                                                    // 500
                                                                                                      // 501
roleAssignmentPattern = {                                                                             // 502
  none : Match.Optional(null),                                                                        // 503
  allAccess: Match.Optional(null),                                                                    // 504
  roleId: Match.Optional(Match.Integer),                                                              // 505
  addPermissions: Match.Optional([Boolean]),                                                          // 506
  removePermissions: Match.Optional([Boolean]),                                                       // 507
};                                                                                                    // 508
                                                                                                      // 509
SandstormDb = function () {                                                                           // 510
  this.collections = {                                                                                // 511
    // Direct access to underlying collections. DEPRECATED.                                           // 512
    //                                                                                                // 513
    // TODO(cleanup): Over time, we will provide methods covering each supported query and remove     // 514
    //   direct access to the collections.                                                            // 515
                                                                                                      // 516
    packages: Packages,                                                                               // 517
    devApps: DevApps,                                                                                 // 518
    userActions: UserActions,                                                                         // 519
    grains: Grains,                                                                                   // 520
    contacts: Contacts,                                                                               // 521
    sessions: Sessions,                                                                               // 522
    signupKeys: SignupKeys,                                                                           // 523
    activityStats: ActivityStats,                                                                     // 524
    deleteStats: DeleteStats,                                                                         // 525
    fileTokens: FileTokens,                                                                           // 526
    apiTokens: ApiTokens,                                                                             // 527
    notifications: Notifications,                                                                     // 528
    statsTokens: StatsTokens,                                                                         // 529
    misc: Misc,                                                                                       // 530
    settings: Settings,                                                                               // 531
                                                                                                      // 532
    // Intentionally omitted:                                                                         // 533
    // - Migrations, since it's used only within this package.                                        // 534
    // - RoleAssignments, since it is deprecated and only used by the migration that eliminated it.   // 535
  };                                                                                                  // 536
};                                                                                                    // 537
                                                                                                      // 538
// TODO(cleanup): These methods should not be defined freestanding and should use collection          // 539
//   objects created in SandstormDb's constructor rather than globals.                                // 540
                                                                                                      // 541
_.extend(SandstormDb.prototype, {                                                                     // 542
  isDemoUser: isDemoUser,                                                                             // 543
  isSignedUp: isSignedUp,                                                                             // 544
  isSignedUpOrDemo: isSignedUpOrDemo,                                                                 // 545
  isUserOverQuota: isUserOverQuota,                                                                   // 546
  isUserExcessivelyOverQuota: isUserExcessivelyOverQuota,                                             // 547
  isAdmin: isAdmin,                                                                                   // 548
  isAdminById: isAdminById,                                                                           // 549
  findAdminUserForToken: findAdminUserForToken,                                                       // 550
  matchWildcardHost: matchWildcardHost,                                                               // 551
  makeWildcardHost: makeWildcardHost,                                                                 // 552
  allowDevAccounts: allowDevAccounts,                                                                 // 553
  roleAssignmentPattern: roleAssignmentPattern,                                                       // 554
});                                                                                                   // 555
                                                                                                      // 556
if (Meteor.isServer) {                                                                                // 557
  SandstormDb.prototype.getWildcardOrigin = getWildcardOrigin;                                        // 558
}                                                                                                     // 559
                                                                                                      // 560
////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);






(function () {

////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                                    //
// packages/sandstorm-db/migrations.js                                                                //
//                                                                                                    //
////////////////////////////////////////////////////////////////////////////////////////////////////////
                                                                                                      //
// Minimal tooling for doing run-at-least-once, ordered migrations.                                   // 1
//                                                                                                    // 2
// Because migrations can experience partial failure and likely have                                  // 3
// side-effects, we should be careful to make sure all migrations are                                 // 4
// idempotent and safe to accidentally run multiple times.                                            // 5
                                                                                                      // 6
var updateLoginStyleToRedirect = function() {                                                         // 7
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;          // 8
  ["google", "github"].forEach(function(serviceName) {                                                // 9
    var config = configurations.findOne({service: serviceName});                                      // 10
    if (config && config.loginStyle !== "redirect") {                                                 // 11
      configurations.update({service: serviceName}, {$set: {loginStyle: "redirect"}});                // 12
    }                                                                                                 // 13
  });                                                                                                 // 14
};                                                                                                    // 15
                                                                                                      // 16
var enableLegacyOAuthProvidersIfNotInSettings = function() {                                          // 17
  // In the before time, Google and Github login were enabled by default.                             // 18
  //                                                                                                  // 19
  // This actually didn't make much sense, required the first user to configure                       // 20
  // OAuth, and had some trust-the-first-user properties that weren't totally                         // 21
  // secure.                                                                                          // 22
  //                                                                                                  // 23
  // Now, we have admin-token, so we wish to disable all logins by default                            // 24
  // (since they need to be configured anyway) but since users may have never                         // 25
  // explicitly told Sandstorm that Google or Github login should be enabled,                         // 26
  // we can't just use the value in the Settings collection, since it might                           // 27
  // never have been set.                                                                             // 28
  //                                                                                                  // 29
  // Thus, if the service is configured but absent from Settings, we should                           // 30
  // explicitly enable it in Settings, and then the rest of the logic can just                        // 31
  // depend on what value is in Settings and default to false without breaking                        // 32
  // user installations.                                                                              // 33
  var configurations = Package["service-configuration"].ServiceConfiguration.configurations;          // 34
  ["google", "github"].forEach(function(serviceName) {                                                // 35
    var config = configurations.findOne({service: serviceName});                                      // 36
    var serviceConfig = Settings.findOne({_id: serviceName});                                         // 37
    if (config && !serviceConfig) {                                                                   // 38
      // Only explicitly enable the login service if:                                                 // 39
      // 1) the service is already configured                                                         // 40
      // 2) there is no sandstorm configuration already present (the user was                         // 41
      //    using the previous default behavior).                                                     // 42
      Settings.insert({_id: serviceName, value: true});                                               // 43
    }                                                                                                 // 44
  });                                                                                                 // 45
};                                                                                                    // 46
                                                                                                      // 47
var denormalizeInviteInfo = function() {                                                              // 48
  // When a user is invited via a signup token, the `signupKey` field of their user table entry       // 49
  // has always been populated to indicate the key they used. This points into the SignupKeys table   // 50
  // which has more information about the key, namely a freeform note entered by the admin when       // 51
  // they created the key. In the case that the email invite form was used, the note has the form     // 52
  // "E-mail invite to <address>".                                                                    // 53
  //                                                                                                  // 54
  // Later, we decided it was useful to indicate in the users table visible to the admin              // 55
  // information about the invite terms. Namely, for email invites we want to show the address        // 56
  // and for others we want to show the note. To make this efficient, fields `signupNote` and         // 57
  // `signupEmail` were added to the users table. We can backfill these values by denormalizing       // 58
  // from the SignupKeys table.                                                                       // 59
                                                                                                      // 60
  Meteor.users.find().forEach(function (user) {                                                       // 61
    if (user.signupKey && (typeof user.signupKey) === "string" && user.signupKey !== "admin") {       // 62
      var signupInfo = SignupKeys.findOne(user.signupKey);                                            // 63
      if (signupInfo && signupInfo.note) {                                                            // 64
        var newFields = { signupNote: signupInfo.note };                                              // 65
                                                                                                      // 66
        var prefix = "E-mail invite to ";                                                             // 67
        if (signupInfo.note.lastIndexOf(prefix) === 0) {                                              // 68
          newFields.signupEmail = signupInfo.note.slice(prefix.length);                               // 69
        }                                                                                             // 70
                                                                                                      // 71
        Meteor.users.update(user._id, {$set: newFields});                                             // 72
      }                                                                                               // 73
    }                                                                                                 // 74
  });                                                                                                 // 75
}                                                                                                     // 76
                                                                                                      // 77
function mergeRoleAssignmentsIntoApiTokens() {                                                        // 78
  RoleAssignments.find().forEach(function(roleAssignment) {                                           // 79
    ApiTokens.insert({                                                                                // 80
      grainId: roleAssignment.grainId,                                                                // 81
      userId: roleAssignment.sharer,                                                                  // 82
      roleAssignment: roleAssignment.roleAssignment,                                                  // 83
      petname: roleAssignment.petname,                                                                // 84
      created: roleAssignment.created,                                                                // 85
      owner: {user: {userId: roleAssignment.recipient,                                                // 86
                     title: roleAssignment.title}},                                                   // 87
    });                                                                                               // 88
  });                                                                                                 // 89
}                                                                                                     // 90
                                                                                                      // 91
function fixOasisStorageUsageStats() {                                                                // 92
  // Before storage quota enforcement was first implemented, there were some bugs with the way the    // 93
  // Oasis backend tracked storage usage, which were not initially noticed since the resulting        // 94
  // number wasn't acutally used for anything.                                                        // 95
  //                                                                                                  // 96
  // With the underlying bugs fixed, simply starting up each grain (supervisor only, not the app)     // 97
  // and then immediately shutting it down should cause the sizes to be recalculated correctly.       // 98
  //                                                                                                  // 99
  // This problem only applies to Oasis.                                                              // 100
                                                                                                      // 101
  if (Meteor.settings.public.quotaEnabled) {                                                          // 102
    var dummyCommand = {                                                                              // 103
      argv: ["foo", "bar"],                                                                           // 104
      environ: []                                                                                     // 105
    };                                                                                                // 106
                                                                                                      // 107
    Grains.find().forEach(function (grain) {                                                          // 108
      try {                                                                                           // 109
        console.log("refreshing grain:", grain._id);                                                  // 110
                                                                                                      // 111
        // We take a highly synchronous approach here so that the logging is understandable.          // 112
        var supervisor = waitPromise(sandstormBackend.startGrain(                                     // 113
            grain.userId, grain._id, grain.packageId, dummyCommand, false, false)).supervisor;        // 114
        try {                                                                                         // 115
          waitPromise(supervisor.shutdown());                                                         // 116
          throw new Error("shutdown() didn't disconnect!");                                           // 117
        } catch (err) {                                                                               // 118
          if (err.kjType !== "disconnected") throw err;                                               // 119
        }                                                                                             // 120
      } catch (err) {                                                                                 // 121
        console.log(err.stack);                                                                       // 122
      }                                                                                               // 123
    });                                                                                               // 124
  }                                                                                                   // 125
}                                                                                                     // 126
                                                                                                      // 127
// This must come after all the functions named within are defined.                                   // 128
// Only append to this list!  Do not modify or remove list entries;                                   // 129
// doing so is likely change the meaning and semantics of user databases.                             // 130
var MIGRATIONS = [                                                                                    // 131
  updateLoginStyleToRedirect,                                                                         // 132
  enableLegacyOAuthProvidersIfNotInSettings,                                                          // 133
  denormalizeInviteInfo,                                                                              // 134
  mergeRoleAssignmentsIntoApiTokens,                                                                  // 135
  fixOasisStorageUsageStats,                                                                          // 136
];                                                                                                    // 137
                                                                                                      // 138
function migrateToLatest() {                                                                          // 139
  var applied = Migrations.findOne({_id: "migrations_applied"});                                      // 140
  var start;                                                                                          // 141
  if (!applied) {                                                                                     // 142
    // Migrations table is not yet seeded with a value.  This means it has                            // 143
    // applied 0 migrations.  Persist this.                                                           // 144
    Migrations.insert({_id: "migrations_applied", value: 0});                                         // 145
    start = 0;                                                                                        // 146
  } else {                                                                                            // 147
    start = applied.value;                                                                            // 148
  }                                                                                                   // 149
  console.log("Migrations applied: " + start + "/" + MIGRATIONS.length);                              // 150
                                                                                                      // 151
  for (var i = start ; i < MIGRATIONS.length ; i++) {                                                 // 152
    // Apply migration i, then record that migration i was successfully run.                          // 153
    console.log("Applying migration " + (i+1));                                                       // 154
    MIGRATIONS[i]();                                                                                  // 155
    Migrations.update({_id: "migrations_applied"}, {$set: {value: i+1}});                             // 156
    console.log("Applied migration " + (i+1));                                                        // 157
  }                                                                                                   // 158
}                                                                                                     // 159
                                                                                                      // 160
// Apply all migrations on startup.                                                                   // 161
Meteor.startup(migrateToLatest);                                                                      // 162
                                                                                                      // 163
////////////////////////////////////////////////////////////////////////////////////////////////////////

}).call(this);


/* Exports */
if (typeof Package === 'undefined') Package = {};
Package['sandstorm-db'] = {
  SandstormDb: SandstormDb
};

})();

//# sourceMappingURL=sandstorm-db.js.map
